use crate::prelude::*;

use anyhow::Result;
use asset::AssetMesh;
use context::prelude::cgmath::Matrix4;

use std::{
    ops::{Index, IndexMut},
    slice::{Iter, IterMut},
};

pub struct Draw {
    meshes: Vec<AssetMesh>,

    model_buffer_callback: Option<Box<dyn Fn(&AssetMesh) + Send + Sync>>,
}

impl Draw {
    pub fn new(meshes: Vec<AssetMesh>) -> Draw {
        Draw {
            meshes,

            model_buffer_callback: None,
        }
    }

    pub fn set_model_buffer_callback<F>(&mut self, callback: F)
    where
        F: Fn(&AssetMesh) + Send + Sync + 'static,
    {
        self.model_buffer_callback = Some(Box::new(callback));
    }

    // Removes a mesh by its name
    pub fn remove_by_name(&mut self, name: &str) -> Option<AssetMesh> {
        self.meshes
            .iter()
            .position(|mesh| match &mesh.name {
                Some(mesh_name) => mesh_name == name,
                None => false,
            })
            .map(|index| self.remove_at(index))
    }

    // Removes a mesh at given index
    pub fn remove_at(&mut self, index: usize) -> AssetMesh {
        self.meshes.remove(index)
    }

    // Add a mesh to this draw
    pub fn push(&mut self, mesh: AssetMesh) {
        if let Some(model_buffer_callback) = &self.model_buffer_callback {
            model_buffer_callback(&mesh);
        }

        self.meshes.push(mesh);
    }

    // Clears all meshes
    pub fn clear(&mut self) {
        self.meshes.clear();
    }

    pub fn contains(&self, name: &str) -> bool {
        self.meshes.iter().any(|mesh| match &mesh.name {
            Some(mesh_name) => mesh_name == name,
            None => false,
        })
    }

    pub fn iter(&self) -> Iter<'_, AssetMesh> {
        self.meshes.iter()
    }

    pub fn iter_mut(&mut self) -> IterMut<'_, AssetMesh> {
        self.meshes.iter_mut()
    }

    pub fn set_transform(&self, transform: Matrix4<f32>) -> Result<()> {
        self.iter().try_for_each(|mesh| {
            mesh.model_buffer()
                .fill(&[(transform * mesh.model_matrix()).into()])
        })
    }

    pub fn len(&self) -> usize {
        self.meshes.len()
    }

    pub fn is_empty(&self) -> bool {
        self.len() == 0
    }

    // pub fn get(&self, index: usize) -> Option<&AssetMesh> {
    //     self.meshes.get(index)
    // }

    // pub fn get_mut(&mut self, index: usize) -> Option<&mut AssetMesh> {
    //     self.meshes.get_mut(index)
    // }
}

impl Clone for Draw {
    fn clone(&self) -> Self {
        Self {
            meshes: self.meshes.clone(),

            model_buffer_callback: None,
        }
    }
}

// only modify the meshes if you know what you are doing
impl IndexMut<usize> for Draw {
    fn index_mut(&mut self, index: usize) -> &mut AssetMesh {
        &mut self.meshes[index]
    }
}

impl Index<usize> for Draw {
    type Output = AssetMesh;

    fn index(&self, index: usize) -> &AssetMesh {
        &self.meshes[index]
    }
}

impl EntityComponent for Draw {
    fn name(&self) -> &str {
        Self::debug_name()
    }
}

impl ComponentDebug for Draw {
    fn debug_name() -> &'static str {
        "Draw"
    }
}
