pub mod rasterizer;
pub mod raytracer;
pub mod shared;

use crate::prelude::*;

use anyhow::Result;
use utilities::prelude::cgmath::Vector3;

use std::sync::Arc;

pub trait RenderingFrontEnd {
    fn render_scale(&self) -> f32;

    // user point of view handle
    fn view_mut(&mut self) -> &mut View;
    fn view(&self) -> &View;
    fn set_center(&mut self, center: Vector3<f32>) -> Result<()>;

    fn screen_to_world(&self, x: u32, y: u32) -> Result<Option<Vector3<f32>>>;

    // scene rendering
    fn process(
        &mut self,
        changed: bool,
        content: Vec<&EntityObject>,
        buffer_recorder: &mut CommandBufferRecorder<'_>,
        images: &TargetMode<Vec<Arc<Image>>>,
        indices: &TargetMode<usize>,
    ) -> Result<()>;

    fn resize(
        &mut self,
        window_width: f32,
        window_height: f32,
        images: &TargetMode<Vec<Arc<Image>>>,
    ) -> Result<()>;

    // light handling
    fn add_light(&mut self, light: Light) -> Result<()>;
    fn remove_light(&mut self, light: Light) -> Result<()>;
    fn clear_lights(&mut self) -> Result<()>;
}

pub trait ExtensionCheck {
    fn check(device: &Arc<Device>) -> bool;
}

pub struct RasterizerDescriptorLayouts;

impl RasterizerDescriptorLayouts {
    pub fn light_descriptor_layout(device: Arc<Device>) -> Result<Arc<DescriptorSetLayout>> {
        DescriptorSetLayout::builder()
            .add_layout_binding(
                0,
                VK_DESCRIPTOR_TYPE_UNIFORM_BUFFER,
                VK_SHADER_STAGE_VERTEX_BIT,
                0,
            )
            .add_layout_binding(
                1,
                VK_DESCRIPTOR_TYPE_COMBINED_IMAGE_SAMPLER,
                VK_SHADER_STAGE_FRAGMENT_BIT,
                0,
            )
            .build(device)
    }

    pub fn deferred_descriptor_layout(device: Arc<Device>) -> Result<Arc<DescriptorSetLayout>> {
        DescriptorSetLayout::builder()
            // position texture
            .add_layout_binding(
                0,
                VK_DESCRIPTOR_TYPE_COMBINED_IMAGE_SAMPLER,
                VK_SHADER_STAGE_FRAGMENT_BIT,
                0,
            )
            // normal texture
            .add_layout_binding(
                1,
                VK_DESCRIPTOR_TYPE_COMBINED_IMAGE_SAMPLER,
                VK_SHADER_STAGE_FRAGMENT_BIT,
                0,
            )
            // albedo texture
            .add_layout_binding(
                2,
                VK_DESCRIPTOR_TYPE_COMBINED_IMAGE_SAMPLER,
                VK_SHADER_STAGE_FRAGMENT_BIT,
                0,
            )
            // scene info
            .add_layout_binding(
                3,
                VK_DESCRIPTOR_TYPE_UNIFORM_BUFFER,
                VK_SHADER_STAGE_FRAGMENT_BIT,
                0,
            )
            .build(device)
    }

    pub fn light_info_descriptor(device: Arc<Device>) -> Result<Arc<DescriptorSetLayout>> {
        DescriptorSetLayout::builder()
            .add_layout_binding(
                0,
                VK_DESCRIPTOR_TYPE_UNIFORM_BUFFER,
                VK_SHADER_STAGE_FRAGMENT_BIT,
                0,
            )
            .build(device)
    }
}

pub struct RaytracerDesciptorLayouts;

impl RaytracerDesciptorLayouts {
    pub fn vertex_type_descriptor_layout(device: Arc<Device>) -> Result<Arc<DescriptorSetLayout>> {
        DescriptorSetLayout::builder()
            // scene data
            .add_layout_binding(
                0,
                VK_DESCRIPTOR_TYPE_STORAGE_BUFFER,
                VK_SHADER_STAGE_COMPUTE_BIT
                    | VK_SHADER_STAGE_CLOSEST_HIT_BIT_KHR
                    | VK_SHADER_STAGE_ANY_HIT_BIT_KHR,
                0,
            )
            .set_flags(0)
            .build(device)
    }

    pub fn light_descriptor_layout(
        device: Arc<Device>,
        max: u32,
    ) -> Result<Arc<DescriptorSetLayout>> {
        DescriptorSetLayout::builder()
            .add_layout_binding(
                0,
                VK_DESCRIPTOR_TYPE_STORAGE_BUFFER,
                VK_SHADER_STAGE_COMPUTE_BIT | VK_SHADER_STAGE_CLOSEST_HIT_BIT_KHR,
                VK_DESCRIPTOR_BINDING_VARIABLE_DESCRIPTOR_COUNT_BIT_EXT
                    | VK_DESCRIPTOR_BINDING_PARTIALLY_BOUND_BIT_EXT,
            )
            .change_descriptor_count(max)
            .build(device)
    }
}
